"""Properties of matrices and vectors."""
from toqito.matrix_props.is_square import is_square
from toqito.matrix_props.is_diagonal import is_diagonal
from toqito.matrix_props.is_normal import is_normal
from toqito.matrix_props.is_symmetric import is_symmetric
from toqito.matrix_props.is_hermitian import is_hermitian
from toqito.matrix_props.is_identity import is_identity
from toqito.matrix_props.is_idempotent import is_idempotent
from toqito.matrix_props.is_positive_semidefinite import is_positive_semidefinite
from toqito.matrix_props.is_positive_definite import is_positive_definite
from toqito.matrix_props.is_density import is_density
from toqito.matrix_props.is_commuting import is_commuting
from toqito.matrix_props.is_projection import is_projection
from toqito.matrix_props.is_unitary import is_unitary
from toqito.matrix_props.majorizes import majorizes
