"""A number of widely-studied quantum channels."""

from toqito.channels.choi import choi
from toqito.channels.dephasing import dephasing
from toqito.channels.depolarizing import depolarizing
from toqito.channels.partial_trace import partial_trace
from toqito.channels.partial_transpose import partial_transpose
from toqito.channels.realignment import realignment
from toqito.channels.reduction import reduction
