"""A number of nonlocal game-related functions for toqito."""
from toqito.nonlocal_games.extended_nonlocal_game import ExtendedNonlocalGame
from toqito.nonlocal_games.nonlocal_game import NonlocalGame
from toqito.nonlocal_games.quantum_hedging import QuantumHedging
from toqito.nonlocal_games.xor_game import XORGame
